CON
  _clkmode = XTAL1|PLL16X
  _xinfreq = 5_000_000
  
CON
  columns  = driver#res_x / 8
  rows     = driver#res_y / font#height
  bcnt     = columns * rows

  rows_raw = (driver#res_y + font#height - 1) / font#height
  bcnt_raw = columns * rows_raw

  dcolour  = driver#dcolour & $FFFF * $00010001

  idle = driver#hv_idle
  mask = driver#hv_mask

  #8, BS, TAB, LF, VT, FF, CR, ESC = 27       
OBJ
  driver: "waitvid.50xH.driver.2048"
    font: "generic8x8-1font"
  
VAR
  long  link[driver#res_m] ,cext                              ' driver mailbox
  long  plte[bcnt_raw / 2]                              ' colour buffer (4n aligned)
  word  scrn[bcnt_raw / 2]                              ' screen buffer (2n aligned)

  byte  xx, yy, page                                      ' cursor position, page mode
  word  flag, colour
  long  sptr
{ 
PUB selftest | x, y, c

  longfill(@plte{0}, dcolour, constant(bcnt_raw / 2))   ' default colour
  wordfill(@scrn{0},   $2020, constant(bcnt_raw / 2))   ' clear screen

  init                                                  ' start driver

  waitcnt(clkfreq*3 + cnt)
  redefine_char($20, @new_spc)                          ' new space character
  waitcnt(clkfreq*2 + cnt)
  redefine_char($20, @old_spc)                          ' reset

' setup plot functionality

  redefine_plot                                         ' 128..143

' draw some dots

  repeat
    repeat bcnt_raw * 5
      plot(||(?x//(columns*2)), ||(y?//(rows*2)), ?c)
    repeat bcnt_raw * 5
      plot(||(?x//(columns*2)), ||(y?//(rows*2)), -1)
}   
DAT
new_spc byte    $FF, $81[6], $FF, $81[3], $FF, $81[3], $FF
old_spc byte    0[16]

PUB putc(c)

  if cext                                               ' colour per position mode
    word[cext][yy * columns + xx] := colour               ' update colour
 
  scrn.byte[bcnt_raw - yy * columns - ++xx] := c          ' pre-increment resolves (bcnt_r -1)
  if xx == columns                                         
    xx := newline                                        ' CR/LF

PRI newline

  if ++yy == rows
    if page                                             ' page/scroll?
      yy := 0
      return
    yy--

'   If we have a colour buffer attached scroll it up. The last active line
'   is unconditionally filled with the current colour. An unused line (i.e.
'   bcnt <> bcnt_r) at the bottom will retain the last clear screen colour.
'
    if cext
      wordmove(cext, @word[cext][columns], constant(bcnt_raw - columns))
      wordfill(@word[cext][constant(bcnt - columns)], colour, columns)
      
    wordmove(@scrn.byte[columns], @scrn{0}, constant((bcnt_raw - columns) / 2))
    if rows_raw == rows
      wordfill(@scrn{0}, $2020, constant(columns / 2))

PUB str(addr)

  repeat strsize(addr)
    out(byte[addr++])

PUB out(c) : succ
'' Output a character
''
''     $00 = NUL   clear screen
''     $01 = SOH   home
''     $08 = BS  * backspace
''     $09 = TAB * tab
''     $0A = LF    set X position (X follows)
''     $0B = VT    set Y position (Y follows)
''     $0C = FF  * clear screen
''     $0D = CR  * return
''     $1B = ESC sequence(s)
''  others = printable characters

  case flag.byte{0}
    $00: case c                                             
           $00..$01,FF:                            
             if c <> $01                           
               if cext
                 wordfill(cext, colour, bcnt_raw)
               wordfill(@scrn{0}, $2020, constant(bcnt_raw / 2))
             xx := yy := 0                           
           $08: xx := (xx - 1) #> 0                  
           TAB: repeat 8 - (xx & 7)                    
                  putc(" ")                         
           $0A..$0B: flag := c << 8 | "="               ' emulate head/tail of ESC=<x><y>
           $0D: xx := newline                            ' CR/LF
           ESC: flag := c                           
           other: putc(c)                            
    ESC: case c
           "s", "S": page := c & $20                    ' page (ESC+s) and scroll mode (ESC+S)
           "=":      succ := constant($0200 | "=")      ' ESC=<x><y>
           "c":      succ := constant($0200 | "c")      ' ESCc<foreground><background>
         flag := succ~
    "=": case flag.byte[1]--
           2..10: flag &= tr(c, @xx, columns, flag >> 11)
           other: flag := tr(c, @yy, rows, TRUE)
    "c": case flag.byte[1]--
           2:{flag |=}sc(c, 1)
           1: flag := sc(c, 0)
           
PRI tr(c, addr, limit, response)

  if c.byte{0} < 255
    byte[addr] := c.byte{0} // limit

  return not response

PRI sc(c, idx)

  colour.byte[idx] := c & mask | idle

PUB setn(n, addr) | base

  if n == 2                                             ' colour ...
    if cext := (addr &= -4) #> 0                        ' colour buffer (4n)
      base := (addr - @base) ~> 2                       ' |
      repeat constant(bcnt_raw / 2)                     ' |
        base[base++] := base[base] & mask | idle        ' clean up sync bits

  link[n] := addr
  repeat
  while link[n]

PUB hex(value, digits)

  value <<= (8 - digits) << 2
  repeat digits
    putc(lookupz((value <-= 4) & %1111 : "0".."9", "A".."F"))

PUB bin(value, digits)

  value <<= 32 - digits
  repeat digits
    putc((value <-= 1) & 1 + "0")

PUB dec(value) | s[4], p

  s{0} := value < 0                                     ' remember sign

  p := @p                                               ' initialise string pointer
  byte[--p] := 0                                        ' terminate string
  
  repeat
    byte[--p] := ||(value // 10) + "0"                  ' |
    value /= 10                                         ' |
  while value                                           ' create decimal representation

  if s{0}                                               ' optionally
    byte[--p] := "-"                                    ' prepend sign

  repeat strsize(p)
    putc(byte[p++])                                     ' emit string


PUB TEXTCOLOR(value)
   colour.byte[1] := value & mask | idle  
'' Added by Jeff Ledger

PUB BACKCOLOR(value)
   colour.byte[0] := value & mask | idle
'' Added by Jeff Ledger

PUB CLS
   'str(string(ESC, "c", %%0003, %%0003, FF))
   out($00)
'' Added By Jeff Ledger
     
PUB init

  longfill(@plte{0}, dcolour, constant(bcnt_raw / 2))   ' default colour
  wordfill(@scrn{0},   $2020, constant(bcnt_raw / 2))   ' clear screen

  link{0} := @scrn.byte[bcnt_raw - columns]             ' initial screen buffer and
  link[1] := font#height << 24 | font.addr              ' font definition
  link[2] := @plte{0}                                   ' colour buffer
  
  return driver.init(-1, @link{0})                      ' video driver and pixel generator

PRI redefine_char(c, buffer) | base

  base := font.addr + c.byte{0}                         ' start location
  repeat font#height
    byte[base] := byte[buffer++]
    base += 256
    
PUB redefine_plot | i, ptr

  ptr := font.addr 
  repeat i from 0 to constant(font#height/2 -1)
    bytemove(ptr + (i << 8) +                                  128,  @sg1, 16)
    bytemove(ptr + (i << 8) + constant(font#height / 2 * 256 + 128), @sg2, 16)
  
DAT
sg1     byte    $00, $0F, $F0, $FF, $00, $0F, $F0, $FF, $00, $0F, $F0, $FF, $00, $0F, $F0, $FF
sg2     byte    $00, $00, $00, $00, $0F, $0F, $0F, $0F, $F0, $F0, $F0, $F0, $FF, $FF, $FF, $FF

PUB plot(x, y, c) | pv, nv

  nv := |< (((y & 1) << 1) | (x & 1))   
  x >>= 1
  y >>= 1
  pv := c_get(x, y) & $0F
  if c == -1
    nv := pv & !nv
  else
    nv := pv |  nv
  c_set(x, y, nv|128)    

' finally set colour

  plte.byte[(y * columns + x) << 1 + 1] := c & driver#hv_mask | driver#hv_idle
  
PRI c_set(x, y, c)

  scrn.byte[bcnt_raw - (y * columns) - ++x] := c

PRI c_get(x, y) : c

  c := scrn.byte[bcnt_raw - (y * columns) - ++x]

DAT